import json
import uuid
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from attr import asdict
from attr import define
from typingref import UNSET

from qtgql.utils.graphql import get_operation_name


@define
class EncodeAble:
    def asdict(self) -> dict:
        return {k: v for k, v in asdict(self).items() if v is not UNSET}


T = TypeVar("T")


@define(kw_only=True)
class QueryPayload(Generic[T], EncodeAble):
    query: str
    operationName: Optional[str] = UNSET
    variables: Optional[Union[dict, T]] = UNSET

    def __attrs_post_init__(self):
        if not self.operationName:
            self.operationName = get_operation_name(self.query)


@define(kw_only=True, repr=False)
class SafeGqlPayload(QueryPayload):
    ...


@define(kw_only=True)
class GqlResult:
    data: dict
    errors: Optional[list[dict]] = None


class GqlEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, (EncodeAble, QueryPayload)):
            return obj.asdict()



from __future__ import annotations

import inspect
from typing import Any
from typing import TYPE_CHECKING

from attrs import define
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.py.compiler.config import QtGqlConfig
    from qtgql.codegen.py.compiler.query import QtGqlOperationDefinition, QtGqlQueriedField
    from qtgql.codegen.py.objecttype import (
        QtGqlEnumDefinition,
        QtGqlInputObjectTypeDefinition,
        QtGqlInterfaceDefinition,
        QtGqlObjectTypeDefinition,
    )

template_env = Environment(loader=PackageLoader("qtgql.codegen.py"), autoescape=select_autoescape())

SCHEMA_TEMPLATE = template_env.get_template("schema.jinja.py")
HANDLERS_TEMPLATE = template_env.get_template("handlers.jinja.py")
CONFIG_TEMPLATE = template_env.get_template("config.jinja.py")


@define
class TemplateContext:
    enums: list[QtGqlEnumDefinition]
    types: list[QtGqlObjectTypeDefinition]
    interfaces: list[QtGqlInterfaceDefinition]
    queries: list[QtGqlOperationDefinition]
    mutations: list[QtGqlOperationDefinition]
    subscriptions: list[QtGqlOperationDefinition]
    input_objects: list[QtGqlInputObjectTypeDefinition]
    config: QtGqlConfig

    @property
    def dependencies(self) -> list[str]:
        def build_import_statement(t: type[Any]) -> str:
            mod = inspect.getmodule(t)
            assert mod
            return f"from {mod.__name__} import {t.__name__}"

        ret = [build_import_statement(scalar) for scalar in self.config.custom_scalars.values()]
        ret.append(build_import_statement(self.config.base_object))
        return ret

    @property
    def custom_scalars(self) -> list[str]:
        return [scalar.__name__ for scalar in self.config.custom_scalars.values()]

    @property
    def base_object_name(self) -> str:
        return self.config.base_object.__name__


def schema_types_template(context: TemplateContext) -> str:
    return SCHEMA_TEMPLATE.render(context=context)


def handlers_template(context: TemplateContext) -> str:
    return HANDLERS_TEMPLATE.render(context=context)


@define
class ConfigContext:
    p_field: QtGqlQueriedField

    @property
    def choices(self):
        if self.p_field.choices:
            return {
                type_name: {
                    selection.name: selection.as_conf_string() or "None" for selection in selections
                }
                for type_name, selections in self.p_field.choices.items()
            }
        else:
            return {}

    @property
    def selections(self) -> dict[str, str]:
        if self.p_field.selections:
            return {
                selection.name: selection.as_conf_string() for selection in self.p_field.selections
            }
        else:
            return {}


def config_template(context: ConfigContext):
    return CONFIG_TEMPLATE.render(context=context)



from __future__ import annotations

from collections import defaultdict
from textwrap import dedent
from typing import NamedTuple
from typing import Optional
from typing import TYPE_CHECKING

import attrs
from typingref import UNSET

from qtgql.codegen.graphql_ref import has_id_selection
from qtgql.codegen.graphql_ref import has_typename_selection
from qtgql.codegen.graphql_ref import inject_id_selection
from qtgql.codegen.graphql_ref import inject_typename_selection
from qtgql.codegen.graphql_ref import is_field_node
from qtgql.codegen.graphql_ref import is_inline_fragment
from qtgql.codegen.py.compiler.template import config_template
from qtgql.codegen.py.compiler.template import ConfigContext
from qtgql.codegen.py.objecttype import QtGqlFieldDefinition
from qtgql.codegen.py.objecttype import QtGqlObjectTypeDefinition

if TYPE_CHECKING:
    from graphql import language as gql_lang

    from qtgql.codegen.introspection import SchemaEvaluator
    from qtgql.codegen.py.objecttype import QtGqlVariableDefinition


def is_type_name_selection(field_node: gql_lang.FieldNode):
    # typename is not a 'real' selection and is handled with special care.
    if field_node.name.value == "__typename":
        return True
    return False


def get_field_from_field_node(
    selection: gql_lang.FieldNode,
    field_type: QtGqlObjectTypeDefinition,
    schema_evaluator: SchemaEvaluator,
    parent_interface_field: Optional[QtGqlQueriedField] = UNSET,
) -> QtGqlQueriedField:
    field_node = is_field_node(selection)
    assert field_node
    return QtGqlQueriedField.from_field(
        field_type.fields_dict[field_node.name.value],
        selection_set=field_node.selection_set,
        schema_evaluator=schema_evaluator,
        parent_interface_field=parent_interface_field,
    )


class UniqueFieldsList(list["QtGqlQueriedField"]):
    def append(self, obj: QtGqlQueriedField) -> None:
        for v in self:
            if v.name == obj.name:
                pass
        super().append(obj)

    def extend(self, other) -> None:
        for v in other:
            self.append(v)


@attrs.define
class QtGqlQueriedField(QtGqlFieldDefinition):
    selections: UniqueFieldsList = attrs.Factory(UniqueFieldsList)
    choices: dict[str, UniqueFieldsList] = attrs.Factory(lambda: defaultdict(UniqueFieldsList))

    @classmethod
    def from_field(
        cls,
        f: QtGqlFieldDefinition,
        selection_set: Optional[gql_lang.SelectionSetNode],
        schema_evaluator: SchemaEvaluator,
        parent_interface_field: Optional[QtGqlQueriedField] = UNSET,
    ) -> QtGqlQueriedField:
        """Main purpose here is to find inner selections of fields, this could
        be an object type, interface, union or a list.

        Any other fields should not have inner selections.
        """
        assert parent_interface_field is not UNSET
        ret = cls(**attrs.asdict(f, recurse=False))
        if not hasattr(selection_set, "selections"):
            return ret
        assert selection_set
        tp = ret.type
        if (
            tp.is_model
        ):  # GraphQL's lists are basically the object beneath them in terms of selections.
            tp = tp.is_model

        tp_is_union = tp.is_union

        # inject id selection for types that supports it. unions are handled below.
        if f.can_select_id and not has_id_selection(selection_set):
            inject_id_selection(selection_set)

        # inject parent interface selections.
        if (tp.is_object_type or tp.is_interface) and parent_interface_field:
            ret.selections.extend(parent_interface_field.selections)

        if tp_is_union:
            for selection in selection_set.selections:
                fragment = is_inline_fragment(selection)
                assert fragment

                type_name = fragment.type_condition.name.value
                concrete = schema_evaluator.get_objecttype_by_name(type_name)
                assert concrete
                if not has_typename_selection(fragment.selection_set):
                    inject_typename_selection(fragment.selection_set)
                if not has_id_selection(fragment.selection_set) and concrete.has_id_field:
                    inject_id_selection(fragment.selection_set)

                for selection_node in fragment.selection_set.selections:
                    field_node = is_field_node(selection_node)
                    assert field_node

                    if not is_type_name_selection(field_node):
                        ret.choices[type_name].append(
                            get_field_from_field_node(
                                field_node,
                                concrete,
                                schema_evaluator,
                                parent_interface_field,
                            ),
                        )

        elif interface_def := tp.is_interface:
            # first get all linear selections.
            for selection in selection_set.selections:
                if not is_inline_fragment(selection):
                    field_node = is_field_node(selection)
                    assert field_node
                    if not is_type_name_selection(field_node):
                        ret.selections.append(
                            get_field_from_field_node(
                                field_node,
                                interface_def,
                                schema_evaluator,
                                parent_interface_field,
                            ),
                        )

            for selection in selection_set.selections:
                if inline_frag := is_inline_fragment(selection):
                    type_name = inline_frag.type_condition.name.value
                    # no need to validate inner types are implementation, graphql-core does this.
                    concrete = schema_evaluator.get_objecttype_by_name(
                        type_name,
                    ) or schema_evaluator.get_interface_by_name(type_name)
                    assert concrete
                    for inner_selection in inline_frag.selection_set.selections:
                        field_node = is_field_node(inner_selection)
                        assert field_node
                        if not is_type_name_selection(field_node):
                            ret.choices[type_name].append(
                                get_field_from_field_node(
                                    field_node,
                                    concrete,
                                    schema_evaluator,
                                    parent_interface_field,
                                ),
                            )

        else:  # object types.
            obj_def = tp.is_object_type
            assert obj_def
            for selection in selection_set.selections:
                field_node = is_field_node(selection)
                assert field_node
                if not is_type_name_selection(field_node):
                    ret.selections.append(
                        get_field_from_field_node(
                            field_node,
                            obj_def,
                            schema_evaluator,
                            parent_interface_field,
                        ),
                    )
        return ret

    def as_conf_string(self) -> str:
        return dedent(
            config_template(
                context=ConfigContext(self),
            ),
        )


class QtGqlOperationDefinition(NamedTuple):
    query: str
    name: str
    operation_type: QtGqlObjectTypeDefinition
    field: QtGqlQueriedField
    directives: list[str] = []
    fragments: list[str] = []
    variables: list[QtGqlVariableDefinition] = []

    @property
    def operation_config(self) -> str:
        return self.field.as_conf_string()



from _decimal import Decimal
from abc import ABC
from abc import abstractmethod
from datetime import date
from datetime import datetime
from datetime import time
from datetime import timezone
from typing import Any
from typing import Generic
from typing import Optional
from typing import Type
from typing import TypeVar

T = TypeVar("T")
T_RAW = TypeVar("T_RAW")
__all__ = ["BaseCustomScalar"]


class BaseCustomScalar(Generic[T, T_RAW], ABC):
    """Class to extend by user defined scalars."""

    __slots__ = "_value"
    GRAPHQL_NAME: str
    """The *real* GraphQL name of the scalar (used by the codegen inspection
    pipeline)."""
    DEFAULT_VALUE: T
    """A place holder graphql query returned null or the field wasn't queried.

    can be used by `from_graphql()`
    """

    def __init__(self, v: Optional[T] = None):
        if not v:
            self._value = self.DEFAULT_VALUE
        else:
            self._value = v

    def parse_value(self) -> T_RAW:
        """When this scalar is used for input types (in variables) this method
        would be called to parse the value to JSON form."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def deserialize(cls, v: Optional[T_RAW] = None) -> "BaseCustomScalar":
        """Deserializes data fetched from graphql, This is useful when you want
        to set a first-of value that will later be used by `to_qt()`.

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def to_qt(self) -> Any:
        """Will be used by the property getter, This is the official value that
        Qt should "understand".

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover

    def __ne__(self, other) -> bool:
        assert isinstance(other, BaseCustomScalar), "can only compare scalar with other scalar."
        return self._value != other._value


class DateTimeScalar(BaseCustomScalar[datetime, str]):
    """An ISO-8601 encoded datetime."""

    GRAPHQL_NAME: str = "DateTime"
    DEFAULT_VALUE = datetime.now()
    FORMAT_STRING = "%H:%M (%m/%d/%Y)"

    def parse_value(self) -> str:
        return self._value.isoformat()

    @classmethod
    def deserialize(cls, v=None) -> "DateTimeScalar":
        if v:
            return cls(datetime.fromisoformat(v).replace(tzinfo=timezone.utc))
        return cls()

    def to_qt(self) -> str:
        return self._value.strftime(self.FORMAT_STRING)


class DateScalar(BaseCustomScalar[date, str]):
    """An ISO-8601 encoded date."""

    GRAPHQL_NAME = "Date"
    DEFAULT_VALUE = date(year=1998, month=8, day=23)

    def parse_value(self) -> str:
        return self._value.isoformat()

    @classmethod
    def deserialize(cls, v=None) -> "DateScalar":
        if v:
            return cls(date.fromisoformat(v))
        return cls()

    def to_qt(self) -> str:
        return self._value.isoformat()


class TimeScalar(BaseCustomScalar[time, str]):
    """an ISO-8601 encoded time."""

    GRAPHQL_NAME = "Time"
    DEFAULT_VALUE = time()

    def parse_value(self) -> str:
        return self._value.isoformat()

    @classmethod
    def deserialize(cls, v: Optional[str] = None) -> "TimeScalar":
        if v:
            return cls(time.fromisoformat(v).replace(tzinfo=timezone.utc))
        return cls()

    def to_qt(self) -> str:
        return self._value.isoformat()


class DecimalScalar(BaseCustomScalar[Decimal, str]):
    """A Decimal value serialized as a string."""

    GRAPHQL_NAME = "Decimal"
    DEFAULT_VALUE = Decimal()

    def parse_value(self) -> str:
        return str(self._value)

    @classmethod
    def deserialize(cls, v: Optional[str] = None) -> "DecimalScalar":
        if v:
            return cls(Decimal(v))
        return cls()

    def to_qt(self) -> str:
        return str(self._value)


CustomScalarMap = dict[str, Type[BaseCustomScalar]]
CUSTOM_SCALARS: CustomScalarMap = {
    DateTimeScalar.GRAPHQL_NAME: DateTimeScalar,
    DecimalScalar.GRAPHQL_NAME: DecimalScalar,
    DateScalar.GRAPHQL_NAME: DateScalar,
    TimeScalar.GRAPHQL_NAME: TimeScalar,
}



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import Generic
from typing import NamedTuple
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar

from PySide6.QtCore import QObject
from PySide6.QtCore import Signal
from PySide6.QtQuick import QQuickItem

from qtgql.codegen.py.runtime.environment import get_gql_env
from qtgql.tools import qproperty
from qtgql.tools import slot

if TYPE_CHECKING:
    from qtgql.gqltransport.client import GqlClientMessage

T_QObject = TypeVar("T_QObject", bound=QObject)


class SelectionConfig(NamedTuple):
    selections: Dict[str, Optional[SelectionConfig]] = {}
    choices: Dict[str, SelectionConfig] = {}


class OperationMetaData(NamedTuple):
    operation_name: str
    selections: Optional[SelectionConfig] = None


T = TypeVar("T")


class BaseOperationHandler(QObject, Generic[T]):
    # class-vars
    ENV_NAME: ClassVar[str]
    OPERATION_METADATA: ClassVar[OperationMetaData]
    _message_template: ClassVar[GqlClientMessage]

    # signals
    dataChanged = Signal()
    completedChanged = Signal()
    operationOnFlightChanged = Signal()
    error = Signal(list)

    setVariables: Callable  # implementation by jinja.

    def __init__(self, parent: Optional[QObject]):
        super().__init__(parent)
        name = self.__class__.__name__
        self.setObjectName(name)
        self._completed: bool = False  # whether the graphql operation completed.
        self._operation_on_the_fly: bool = False
        self._data: Optional[T] = None
        self.environment = get_gql_env(self.ENV_NAME)
        self._variables: dict = {}
        self._consumer: Optional[QmlOperationConsumerABC] = None

    def consume(self, consumer: QmlOperationConsumerABC) -> None:
        self._consumer = consumer
        self.dataChanged.connect(consumer.handlerDataChanged)

    def loose(self) -> None:
        """Releases retention from all children, real implementation is
        generated."""
        raise NotImplementedError

    @slot
    def fetch(self) -> None:
        if not self._operation_on_the_fly and not self._completed:
            self.set_operation_on_flight(True)
            self.environment.client.execute(self)  # type: ignore

    @slot
    def refetch(self) -> None:
        self.set_completed(False)
        self.fetch()

    @property
    def message(self) -> GqlClientMessage:
        self._message_template.payload.variables = self._variables
        return self._message_template

    @qproperty(QObject, notify=dataChanged)
    def data(self) -> Optional[T]:
        return self._data

    def set_completed(self, v: bool) -> None:
        self._completed = v
        self.set_operation_on_flight(False)
        self.completedChanged.emit()

    @qproperty(bool, notify=completedChanged)
    def completed(self):
        return self._completed

    def set_operation_on_flight(self, v: bool) -> None:
        self._operation_on_the_fly = v
        self.operationOnFlightChanged.emit()

    @qproperty(bool, notify=operationOnFlightChanged)
    def operationOnFlight(self) -> bool:
        return self._operation_on_the_fly

    def on_data(self, message: dict) -> None:  # pragma: no cover
        # real is on derived class.
        raise NotImplementedError

    def on_completed(self) -> None:
        # https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md#complete
        self.set_completed(True)

    def on_error(self, message: list[dict[str, Any]]) -> None:
        # https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md#error
        self.set_completed(True)
        self.error.emit(message)

    @slot
    def dispose(self) -> None:
        """Get rid of ownership on related data and set data to null."""
        self.loose()

    def deleteLater(self) -> None:
        self.dispose()
        super().deleteLater()


class BaseQueryHandler(BaseOperationHandler[T]):
    """Each handler will be exposed to QML and."""

    ...


class BaseSubscriptionHandler(BaseOperationHandler[T]):
    ...


class BaseMutationHandler(BaseOperationHandler[T]):
    @slot
    def commit(self) -> None:
        # ATM this is just an alias, we might add some functionality here later like optimistic updates.
        self.refetch()


class QmlOperationConsumerABC(QQuickItem, Generic[T]):
    # signals
    operationNameChanged = Signal()
    autofetchChanged = Signal()
    handlerDataChanged = Signal()

    def __init__(self, parent: Optional[QQuickItem] = None):
        super().__init__(parent)
        self._autofetch: bool = False
        self._handler: BaseOperationHandler[T] = self._get_handler()
        self._handler.consume(self)
        self.destroyed.connect(self._handler.dispose)  # type: ignore

    @qproperty(type=BaseOperationHandler, constant=True)
    def handler(self) -> BaseOperationHandler[T]:
        return self._handler

    def handlerData(self):  # property
        # real implementation in template.
        raise NotImplementedError

    @slot
    def set_autofetch(self, v: bool) -> None:
        self._autofetch = v
        self.autofetchChanged.emit()

    @qproperty(type=bool, notify=autofetchChanged, fset=set_autofetch)
    def autofetch(self) -> bool:
        return self._autofetch

    def componentComplete(self) -> None:
        if self._autofetch:
            self._handler.fetch()
        super().componentComplete()

    @classmethod
    def _get_handler(cls) -> BaseOperationHandler[T]:
        # real implementation is generated.
        raise NotImplementedError



from __future__ import annotations

import warnings
from functools import cached_property
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypedDict
from typing import Union

import graphql
from graphql import language as gql_lang
from graphql import OperationType
from graphql.language import visitor
from graphql.type import definition as gql_def

from qtgql.codegen.graphql_ref import is_enum_definition
from qtgql.codegen.graphql_ref import is_input_definition
from qtgql.codegen.graphql_ref import is_interface_definition
from qtgql.codegen.graphql_ref import is_list_definition
from qtgql.codegen.graphql_ref import is_named_type_node
from qtgql.codegen.graphql_ref import is_non_null_definition
from qtgql.codegen.graphql_ref import is_nonnull_node
from qtgql.codegen.graphql_ref import is_object_definition
from qtgql.codegen.graphql_ref import is_operation_def_node
from qtgql.codegen.graphql_ref import is_scalar_definition
from qtgql.codegen.graphql_ref import is_union_definition
from qtgql.codegen.py.compiler.builtin_scalars import BuiltinScalars
from qtgql.codegen.py.compiler.operation import QtGqlOperationDefinition
from qtgql.codegen.py.compiler.operation import QtGqlQueriedField
from qtgql.codegen.py.compiler.template import handlers_template
from qtgql.codegen.py.compiler.template import schema_types_template
from qtgql.codegen.py.compiler.template import TemplateContext
from qtgql.codegen.py.objecttype import EnumMap
from qtgql.codegen.py.objecttype import EnumValue
from qtgql.codegen.py.objecttype import GqlTypeHinter
from qtgql.codegen.py.objecttype import QtGqlEnumDefinition
from qtgql.codegen.py.objecttype import QtGqlFieldDefinition
from qtgql.codegen.py.objecttype import QtGqlInputFieldDefinition
from qtgql.codegen.py.objecttype import QtGqlInputObjectTypeDefinition
from qtgql.codegen.py.objecttype import QtGqlInterfaceDefinition
from qtgql.codegen.py.objecttype import QtGqlObjectTypeDefinition
from qtgql.codegen.py.objecttype import QtGqlVariableDefinition
from qtgql.codegen.utils import anti_forward_ref
from qtgql.exceptions import QtGqlException

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.py.compiler.config import QtGqlConfig

introspection_query = graphql.get_introspection_query(descriptions=True)


class GeneratedNamespace(TypedDict):
    handlers: str
    objecttypes: str


class QtGqlVisitor(visitor.Visitor):
    """Creates handlers for root operations.

    Also injects id field to types that have id field but not explicitly
    queried it.
    """

    def __init__(self, evaluator: SchemaEvaluator):
        super().__init__()
        self.query_handlers: dict[str, QtGqlOperationDefinition] = {}
        self.mutation_handlers: dict[str, QtGqlOperationDefinition] = {}
        self.subscription_handlers: dict[str, QtGqlOperationDefinition] = {}
        self.evaluator = evaluator

    def _get_root_type_field(
        self,
        root_type: QtGqlObjectTypeDefinition,
        gql_field: gql_lang.FieldNode,
    ) -> QtGqlQueriedField:
        return QtGqlQueriedField.from_field(
            root_type.fields_dict[gql_field.name.value],
            gql_field.selection_set,
            self.evaluator,
            parent_interface_field=None,
        )

    def _parse_variable_definition(
        self,
        var: gql_lang.VariableDefinitionNode,
    ) -> QtGqlVariableDefinition:
        return QtGqlVariableDefinition(
            name=var.variable.name.value,
            type=self.evaluator.get_type(var.type),
            type_map=self.evaluator._objecttypes_def_map,
            scalars=self.evaluator.config.custom_scalars,
            enums=self.evaluator._enums_def_map,
            default_value=var.default_value,
        )

    def enter_operation_definition(self, node, key, parent, path, ancestors) -> None:
        if operation := is_operation_def_node(node):
            if operation.operation in (
                OperationType.QUERY,
                OperationType.MUTATION,
                OperationType.SUBSCRIPTION,
            ):
                root_field: gql_lang.FieldNode = operation.selection_set.selections[0]  # type: ignore
                assert operation.name, "QtGql enforces operations to have names."
                op_name = operation.name.value
                operation_vars: list[QtGqlVariableDefinition] = []
                if variables_def := operation.variable_definitions:
                    for var in variables_def:
                        operation_vars.append(self._parse_variable_definition(var))
                if operation.operation is OperationType.QUERY:
                    assert self.evaluator._query_type
                    root_qtgql_field = self._get_root_type_field(
                        self.evaluator._query_type,
                        root_field,
                    )
                    operation_definition = QtGqlOperationDefinition(
                        query=graphql.print_ast(node),
                        operation_type=self.evaluator._query_type,
                        name=op_name,
                        field=root_qtgql_field,
                        directives=node.directives,
                        variables=operation_vars,
                    )
                    self.query_handlers[op_name] = operation_definition
                elif operation.operation is OperationType.MUTATION:
                    assert (
                        self.evaluator._mutation_type
                    ), "You don't have any mutations on your schema"
                    root_qtgql_field = self._get_root_type_field(
                        self.evaluator._mutation_type,
                        root_field,
                    )
                    operation_definition = QtGqlOperationDefinition(
                        query=graphql.print_ast(node),
                        operation_type=self.evaluator._mutation_type,
                        name=op_name,
                        field=root_qtgql_field,
                        directives=node.directives,
                        variables=operation_vars,
                    )
                    self.mutation_handlers[op_name] = operation_definition

                elif operation.operation is OperationType.SUBSCRIPTION:
                    assert (
                        self.evaluator._subscription_type
                    ), "You don't have a subscription type on your schema"
                    root_qtgql_field = self._get_root_type_field(
                        self.evaluator._subscription_type,
                        root_field,
                    )
                    operation_definition = QtGqlOperationDefinition(
                        query=graphql.print_ast(node),
                        operation_type=self.evaluator._subscription_type,
                        name=op_name,
                        field=root_qtgql_field,
                        directives=node.directives,
                        variables=operation_vars,
                    )
                    self.subscription_handlers[op_name] = operation_definition


class SchemaEvaluator:
    def __init__(self, config: QtGqlConfig):
        self.config = config
        self._objecttypes_def_map: dict[str, QtGqlObjectTypeDefinition] = {}
        self._enums_def_map: EnumMap = {}
        self._input_objects_def_map: dict[str, QtGqlInputObjectTypeDefinition] = {}
        self._interfaces_map: dict[str, QtGqlInterfaceDefinition] = {}
        self._query_handlers: dict[str, QtGqlOperationDefinition] = {}
        self._mutation_handlers: dict[str, QtGqlOperationDefinition] = {}
        self._subscription_handlers: dict[str, QtGqlOperationDefinition] = {}
        self._query_type: Optional[QtGqlObjectTypeDefinition] = None
        self._mutation_type: Optional[QtGqlObjectTypeDefinition] = None
        self._subscription_type: Optional[QtGqlObjectTypeDefinition] = None

    def get_interface_by_name(self, name: str) -> Optional[QtGqlInterfaceDefinition]:
        return self._interfaces_map.get(name, None)

    def get_objecttype_by_name(self, name: str) -> Optional[QtGqlObjectTypeDefinition]:
        return self._objecttypes_def_map.get(name, None)

    def get_type(self, node: gql_lang.TypeNode) -> GqlTypeHinter:
        if nonnull := is_nonnull_node(node):
            return self._evaluate_field_type(
                graphql.type.GraphQLNonNull(
                    self.schema_definition.get_type(nonnull.type.name.value),  # type: ignore
                ),
            )

        if named_type := is_named_type_node(node):
            gql_concrete = self.schema_definition.get_type(named_type.name.value)
            assert gql_concrete
            return self._evaluate_field_type(gql_concrete)
        raise NotImplementedError(node, "Type is not supported as a variable ATM")

    @cached_property
    def schema_definition(self) -> graphql.GraphQLSchema:
        with (self.config.graphql_dir / "schema.graphql").open() as f:
            return graphql.build_schema(f.read())

    @cached_property
    def root_types(self) -> List[Optional[gql_def.GraphQLObjectType]]:
        return [
            self.schema_definition.get_root_type(OperationType.QUERY),
            self.schema_definition.get_root_type(OperationType.MUTATION),
            self.schema_definition.get_root_type(OperationType.SUBSCRIPTION),
        ]

    def _evaluate_field_type(self, t: gql_def.GraphQLType) -> GqlTypeHinter:
        # even though every type in qtgql has a default constructor,
        # hence there is no "real" non-null values
        # we store it as optional for generating nullable checks only for what's required.
        ret = None
        is_optional = True
        if non_null := is_non_null_definition(t):
            t = non_null.of_type
            is_optional = False

        if list_def := is_list_definition(t):
            ret = GqlTypeHinter(
                type=list,
                of_type=(self._evaluate_field_type(list_def.of_type),),
                scalars=self.config.custom_scalars,
            )
        elif scalar_def := is_scalar_definition(t):
            if builtin_scalar := BuiltinScalars.by_graphql_name(scalar_def.name):
                ret = GqlTypeHinter(type=builtin_scalar, scalars=self.config.custom_scalars)
            else:
                ret = GqlTypeHinter(
                    type=self.config.custom_scalars[scalar_def.name],
                    scalars=self.config.custom_scalars,
                )
        elif enum_def := is_enum_definition(t):
            ret = GqlTypeHinter(
                type=anti_forward_ref(name=enum_def.name, type_map=self._enums_def_map),
                scalars=self.config.custom_scalars,
            )
        elif obj_def := is_object_definition(t):
            ret = GqlTypeHinter(
                type=anti_forward_ref(name=obj_def.name, type_map=self._objecttypes_def_map),
                scalars=self.config.custom_scalars,
            )
        elif union_def := is_union_definition(t):
            ret = GqlTypeHinter(
                type=Union,
                of_type=tuple(
                    GqlTypeHinter(
                        type=anti_forward_ref(
                            name=possible.name,
                            type_map=self._objecttypes_def_map,
                        ),
                        scalars=self.config.custom_scalars,
                    )
                    for possible in self.schema_definition.get_possible_types(union_def)
                ),
                scalars=self.config.custom_scalars,
            )
        elif input_def := is_input_definition(t):
            concrete = self.schema_definition.get_type(input_def.name)
            assert isinstance(concrete, gql_def.GraphQLInputObjectType)
            ret = GqlTypeHinter(
                type=self._evaluate_input_type(input_def),
                scalars=self.config.custom_scalars,
            )

        elif interface_def := is_interface_definition(t):
            ret = GqlTypeHinter(
                type=self._evaluate_interface_type(interface_def),
                of_type=(),
                scalars=self.config.custom_scalars,
            )
        if not ret:  # pragma: no cover
            raise NotImplementedError(f"type {t} not supported yet")

        if is_optional:
            return GqlTypeHinter(type=Optional, of_type=(ret,), scalars=self.config.custom_scalars)
        return ret

    def _evaluate_field(self, name: str, field: gql_def.GraphQLField) -> QtGqlFieldDefinition:
        return QtGqlFieldDefinition(
            type=self._evaluate_field_type(field.type),
            name=name,
            type_map=self._objecttypes_def_map,
            scalars=self.config.custom_scalars,
            enums=self._enums_def_map,
            description=field.description,
        )

    def _evaluate_input_field(
        self,
        name: str,
        field: gql_def.GraphQLInputField,
    ) -> QtGqlInputFieldDefinition:
        return QtGqlInputFieldDefinition(
            type=self._evaluate_field_type(field.type),
            name=name,
            type_map=self._objecttypes_def_map,
            scalars=self.config.custom_scalars,
            enums=self._enums_def_map,
            description=field.description,
        )

    def _evaluate_object_type(
        self,
        type_: gql_def.GraphQLObjectType,
    ) -> Optional[QtGqlObjectTypeDefinition]:
        t_name: str = type_.name
        if evaluated := self._objecttypes_def_map.get(t_name, None):
            return evaluated
        if type_ not in self.root_types:
            try:
                id_field = type_.fields["id"]
                if nonull := is_non_null_definition(id_field.type):
                    id_scalar = graphql.type.scalars.GraphQLID
                    if nonull.of_type is not id_scalar:
                        raise QtGqlException(
                            f"id field type must be of the {id_scalar} scalar!"
                            f"\n Got: {nonull.of_type}",
                        )
                else:
                    warnings.warn(
                        stacklevel=2,
                        message="It is best practice to have id field of type ID!"
                        f"\ntype {type_} has: {id_field}",
                    )
            except KeyError:
                warnings.warn(
                    stacklevel=2,
                    message="QtGql enforces types to have ID field"
                    f"type {type_} does not not define an id field.\n"
                    f"fields: {type_.fields}",
                )
        implements = [self._evaluate_interface_type(interface) for interface in type_.interfaces]

        ret = QtGqlObjectTypeDefinition(
            name=t_name,
            implements=implements,
            docstring=type_.description,
            fields_dict={
                name: self._evaluate_field(name, field) for name, field in type_.fields.items()
            },
        )
        assert ret not in self.root_types
        self._objecttypes_def_map[ret.name] = ret
        for interface in type_.interfaces:
            qtgql_interface = self._evaluate_interface_type(interface)
            qtgql_interface.implementations[type_.name] = ret
        return ret

    def _evaluate_input_type(
        self,
        type_: gql_def.GraphQLInputObjectType,
    ) -> QtGqlInputObjectTypeDefinition:
        ret = self._input_objects_def_map.get(type_.name, None)
        if not ret:
            ret = QtGqlInputObjectTypeDefinition(
                name=type_.name,
                docstring=type_.description,
                fields_dict={
                    name: self._evaluate_input_field(name, field)
                    for name, field in type_.fields.items()
                },
            )
            self._input_objects_def_map[ret.name] = ret
        return ret

    def _evaluate_interface_type(
        self,
        interface: gql_def.GraphQLInterfaceType,
    ) -> QtGqlInterfaceDefinition:
        if ret := self._interfaces_map.get(interface.name, None):
            return ret
        implements = [self._evaluate_interface_type(base) for base in interface.interfaces]
        ret = QtGqlInterfaceDefinition(
            name=interface.name,
            implements=implements,
            docstring=interface.description,
            fields_dict={
                name: self._evaluate_field(name, field) for name, field in interface.fields.items()
            },
        )
        self._interfaces_map[ret.name] = ret
        return ret

    def _evaluate_enum(self, enum: gql_def.GraphQLEnumType) -> Optional[QtGqlEnumDefinition]:
        name: str = enum.name

        if definition := self._enums_def_map.get(name, None):
            return definition

        ret = QtGqlEnumDefinition(
            name=name,
            members=[
                EnumValue(name=name, description=val.description or "")
                for name, val in enum.values.items()
            ],
        )

        self._enums_def_map[name] = ret
        return ret

    def parse_schema_concretes(self) -> None:
        for name, type_ in self.schema_definition.type_map.items():
            if name.startswith("__"):
                continue
            if object_definition := is_object_definition(type_):
                if object_type := self._evaluate_object_type(object_definition):
                    if object_definition is self.schema_definition.query_type:
                        self._query_type = object_type
                    elif object_definition is self.schema_definition.mutation_type:
                        self._mutation_type = object_type
                    elif object_definition is self.schema_definition.subscription_type:
                        self._subscription_type = object_type

            elif enum_def := is_enum_definition(type_):
                if enum := self._evaluate_enum(enum_def):
                    self._enums_def_map[enum.name] = enum

    def parse_operations(self) -> None:
        with (self.config.graphql_dir / "operations.graphql").open() as f:
            operations = graphql.parse(f.read())

        # validate the operation against the static schema
        if errors := graphql.validate(self.schema_definition, operations):
            raise QtGqlException([error.formatted for error in errors])

        # get QtGql fields from the query AST and inject ID field.
        operation_miner = QtGqlVisitor(self)
        visitor.visit(operations, operation_miner)
        self._query_handlers.update(operation_miner.query_handlers)
        self._mutation_handlers.update(operation_miner.mutation_handlers)
        self._subscription_handlers.update(operation_miner.subscription_handlers)

    def dumps(self) -> GeneratedNamespace:
        """:return: The generated modules as a string."""
        self.parse_schema_concretes()
        self.parse_operations()
        context = TemplateContext(
            enums=list(self._enums_def_map.values()),
            types=[
                t
                for name, t in self._objecttypes_def_map.items()
                if name not in BuiltinScalars.keys()
            ],
            queries=list(self._query_handlers.values()),
            interfaces=list(self._interfaces_map.values()),
            mutations=list(self._mutation_handlers.values()),
            subscriptions=list(self._subscription_handlers.values()),
            input_objects=list(self._input_objects_def_map.values()),
            config=self.config,
        )
        return GeneratedNamespace(
            handlers=handlers_template(context),
            objecttypes=schema_types_template(context),
        )

    def dump(self):
        """:param file: Path to the directory the codegen would dump to."""
        for fname, content in self.dumps().items():
            with (self.config.graphql_dir / (fname + ".py")).open("w") as fh:
                fh.write(content)



import itertools
from typing import Any
from typing import Callable
from typing import get_type_hints
from typing import TypeVar

from PySide6 import QtCore
from typingref import TypeHinter


def get_combos(stripped_optionals: list[TypeHinter], concretes: list[TypeHinter]):
    required_annotations = tuple([th.type for th in concretes])
    combos = []
    for i in range(len(stripped_optionals) + 1):
        for subset in itertools.combinations([th.type for th in stripped_optionals], i):
            combos.append(subset + required_annotations)
    return combos


T_Callable = TypeVar("T_Callable", bound=Callable)


def slot(func: T_Callable) -> T_Callable:
    def wrapper(func) -> T_Callable:
        anots: dict = get_type_hints(func)
        return_ = TypeHinter.from_annotations(anots.pop("return", None))
        if return_.type is type(None):
            ret_type = None
        elif return_.type is Any:
            ret_type = "QVariant"
        else:
            ret_type = return_.type

        args = [TypeHinter.from_annotations(th) for th in anots.values()]

        stripped_optionals = [th.of_type[0] for th in args if th.is_optional()]
        concretes = [th for th in args if not th.is_optional()]

        if stripped_optionals:
            combos = get_combos(stripped_optionals, concretes)
            for combo in combos:
                func = QtCore.Slot(*combo, result=ret_type)(func)
            return func
        return QtCore.Slot(*[th.type for th in concretes], result=ret_type)(func)

    return wrapper(func)



from __future__ import annotations

from typing import Protocol
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.gqltransport.client import HandlerProto


class NetworkLayerProto(Protocol):
    def execute(self, handler: HandlerProto) -> None:
        """accepts a handler and expected to call the handler's `on_data` /
        `on_error` / 'on_completed' when the operation is completed."""


class QtGqlEnvironment:
    """Encapsulates a schema interaction.

    The schema **must** be coherent with the schema use by the code
    generator. This class is used by the codegen and not to be
    instantiated directly.
    """

    def __init__(self, client: NetworkLayerProto, name: str):
        """
        :param client: The network layer for communicated the GraphQL server,
        all the generated handlers for this environment  would use this layer.
        :param name: This would be used to retrieve this environment from the env map by
        the generated handlers based on the configurations.
        """
        self.client = client
        self.name = name


_ENV_MAP: dict[str, QtGqlEnvironment] = {}
"""In the future this would be usefully if you want to use different schemas,
maybe with a directive."""


def set_gql_env(env: QtGqlEnvironment) -> None:
    _ENV_MAP[env.name] = env


def get_gql_env(name: str) -> QtGqlEnvironment:
    return _ENV_MAP[name]



from __future__ import annotations

import contextlib
from typing import ClassVar
from typing import Generic
from typing import NamedTuple
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar

from PySide6.QtCore import QAbstractListModel
from PySide6.QtCore import QByteArray
from PySide6.QtCore import QObject
from PySide6.QtCore import Qt
from PySide6.QtCore import Signal
from PySide6.QtCore import Slot

if TYPE_CHECKING:
    from typing_extensions import Self

    from qtgql.codegen.py.runtime.queryhandler import OperationMetaData, SelectionConfig
from qtgql.tools import qproperty, slot

__all__ = ["QGraphQListModel", "get_base_graphql_object"]


class QGraphQLInputObjectABC(QObject):
    def asdict(self) -> dict:
        raise NotImplementedError


class _BaseQGraphQLObject(QObject):
    TYPE_NAME: ClassVar[str]
    __singleton__: Self

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)

    @qproperty(str, constant=True)
    def typename(self) -> str:
        return self.TYPE_NAME

    @classmethod
    def from_dict(
        cls,
        parent: QObject,
        data: dict,
        config: SelectionConfig,
        metadata: OperationMetaData,
    ):
        """Creates a new instance from GraphQL raw data."""
        raise NotImplementedError

    def update(self, data, config: SelectionConfig) -> None:
        """updates a node based on new GraphQL data."""
        raise NotImplementedError

    def loose(self, metadata: OperationMetaData) -> None:
        """releases all child objects if exists.

        note that this method would be useful only if the object (or one
        of its children) has an id and a reference in the store,
        otherwise the pointer to this object is release and this object
        would be deleted.
        """
        raise NotImplementedError

    @classmethod
    def default_instance(cls) -> Self:
        # used for default values.
        try:
            return cls.__singleton__  # type: ignore
        except AttributeError:
            cls.__singleton__ = cls()
            return cls.__singleton__


class _BaseQGraphQLObjectWithID(_BaseQGraphQLObject):
    _id: str

    __store__: ClassVar[QGraphQLObjectStore[Self]]

    def __init_subclass__(cls, **kwargs):
        cls.__store__ = QGraphQLObjectStore()


T_BaseQGraphQLObjectWithID = TypeVar("T_BaseQGraphQLObjectWithID", bound=_BaseQGraphQLObjectWithID)


def compare_node(node: T_BaseQGraphQLObjectWithID, id_: str, typename: Optional[str] = None):
    if node._id == id_:
        if typename and node.TYPE_NAME != typename:
            return False
        return True


class NodeRecord(NamedTuple):
    node: _BaseQGraphQLObjectWithID
    retainers: set[str]  # set of operation names.

    def retain(self, operation_name: str) -> Self:
        self.retainers.add(operation_name)
        return self


class QGraphQLObjectStore(Generic[T_BaseQGraphQLObjectWithID]):
    def __init__(self) -> None:
        self._data: dict[str, NodeRecord] = {}

    def get_node(self, id_: str) -> Optional[_BaseQGraphQLObject]:
        assert id_
        if found := self._data.get(id_, None):
            return found.node

    def add_record(self, record: NodeRecord):
        assert record.node._id
        self._data[record.node._id] = record

    def loose(self, node: T_BaseQGraphQLObjectWithID, operation_name: str) -> None:
        assert node._id
        with contextlib.suppress(
            KeyError,
        ):  # This node was already deleted, we can safely ignore it
            record = self._data[node._id]
            record.retainers.remove(operation_name)
            if not record.retainers:
                self._data.pop(node._id)
                node.deleteLater()  # we can delete it now since it has no retainers.


class QGraphQListModel(QAbstractListModel, Generic[T_BaseQGraphQLObjectWithID]):
    OBJECT_ROLE = Qt.ItemDataRole.UserRole + 1
    _role_names = {OBJECT_ROLE: QByteArray("object")}  # type: ignore
    currentIndexChanged = Signal()

    def __init__(
        self,
        parent: Optional[QObject],
        data: list[T_BaseQGraphQLObjectWithID],
    ):
        super().__init__(parent)
        self._data = data
        self._current_index: int = 0

    @slot
    def set_current_index(self, i: int) -> None:
        self._current_index = i
        self.currentIndexChanged.emit()

    @qproperty(int, notify=currentIndexChanged, fset=set_current_index)
    def currentIndex(self) -> int:
        return self._current_index

    @qproperty(QObject, notify=currentIndexChanged)  # type: ignore
    def currentObject(self) -> Optional[T_BaseQGraphQLObjectWithID]:
        return self._data[self._current_index]

    def rowCount(self, *args, **kwargs) -> int:
        return len(self._data)

    def roleNames(self) -> dict:
        return self._role_names  # type: ignore

    def data(self, index, role=...) -> Optional[T_BaseQGraphQLObjectWithID]:
        if index.row() < len(self._data) and index.isValid():
            if role == self.OBJECT_ROLE:
                return self._data[index.row()]
            raise NotImplementedError(
                f"role {role} is not a valid role for {self.__class__.__name__}",
            )

    def append(self, node: T_BaseQGraphQLObjectWithID) -> None:
        count = self.rowCount()
        self.beginInsertRows(self.index(count), count, count)
        self._data.append(node)
        self.endInsertRows()

    @slot
    def pop(self, index: Optional[int] = None) -> None:
        index = -1 if index is None else index
        real_index = index if index > -1 else self.rowCount()
        self.beginRemoveRows(self.index(index - 1).parent(), real_index, real_index)
        self._data.pop(index)
        self.endRemoveRows()

    @slot
    def clear(self) -> None:
        if self._data:
            self.beginRemoveRows(self.index(0).parent(), 0, self.rowCount())
            self._data = []
            self.endRemoveRows()

    @Slot(int, QObject)
    def insert(self, index: int, v: T_BaseQGraphQLObjectWithID):
        model_index = self.index(index)
        if index <= self.rowCount() - 1:
            self.beginInsertRows(model_index, index, index)
            self._data[index] = v
            self.endInsertRows()
        else:
            model_index = self.index(self.rowCount() - 1)
            index = self.rowCount()
            self.beginInsertRows(model_index, index, index)
            self._data.append(v)
            self.endInsertRows()

    def update(self, data: list[dict], node_selection: SelectionConfig) -> None:
        raise NotImplementedError

    def removeRows(self, row: int, count: int, parent=None) -> bool:
        # check that this is a valid removal.
        if row + count <= self.rowCount():
            self.beginRemoveRows(self.index(0).parent(), row, count)
            end = self._data[row + count :]
            start = self._data[:row]
            self._data = start + end
            self.endRemoveRows()
            return True
        return False  # pragma: no cover


def get_base_graphql_object(name: str) -> type[_BaseQGraphQLObject]:
    """
    :param name: valid attribute name (used by codegen to import it).
    :returns: A type to be extended by all generated types.
    """
    return type(name, (_BaseQGraphQLObject,), {"__store__": QGraphQLObjectStore()})  # type: ignore


BaseGraphQLObject = get_base_graphql_object("BaseGraphQLObject")

T_BaseModel = TypeVar("T_BaseModel", bound=QGraphQListModel)



from __future__ import annotations

import types
from functools import cached_property
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

import attrs
from PySide6.QtCore import QAbstractListModel
from PySide6.QtCore import QByteArray
from PySide6.QtCore import QEventLoop
from PySide6.QtCore import QModelIndex
from PySide6.QtCore import Qt
from PySide6.QtCore import Signal
from typing_extensions import dataclass_transform
from typing_extensions import Self
from typingref import TypeHinter
from typingref import UNSET

from qtgql.exceptions import QtGqlException
from qtgql.tools import slot

IS_GQL = "is_gql"
IS_ROLE = "is_role"
OLD_NAME = "old_name"

__all__ = ["role", "get_base_type", "GenericModel", "RoleDoesNotExist", "asdict"]


def asdict(t: _BaseType) -> dict:
    return attrs.asdict(t)  # type: ignore


def role(
    default=None,
    factory=UNSET,
):
    """role is optional by default."""
    if factory is not UNSET:
        default = UNSET
    return attrs.field(
        default=attrs.NOTHING if default is UNSET else default,
        factory=None if factory is UNSET else factory,
        metadata={IS_ROLE: True},
    )


def field_is(key: str, field):
    metadata = getattr(field, "metadata", {})
    return metadata.get(key, False)


T_V = TypeVar("T_V")


@dataclass_transform(
    field_specifiers=(role,),
    kw_only_default=True,
)
class BaseTypeMeta(type):
    def __new__(mcs, name, bases, ns):
        super_new = super().__new__

        # exclude BaseType itself
        parents = [b for b in bases if isinstance(b, BaseTypeMeta)]
        cls = super_new(mcs, name, bases, ns)
        if not parents:
            # remove annotations to avoid global reference errors.
            cls.__annotations__ = {}
            return cls
        if not attrs.has(cls):
            assert cls not in cls.__types_map__.values()
            cls: type[_BaseType] = attrs.define(cls, slots=True, kw_only=True)
            cls.__types_map__[cls.__name__] = cls
            roles = RoleMapper()
            for role_num, field in enumerate(attrs.fields(cls), int(Qt.ItemDataRole.UserRole)):  # type: ignore
                # assign role and check if not exists
                if field_is(IS_ROLE, field):
                    role_ = Role(
                        name=field.name,
                        num=role_num,
                        type_map=cls.__types_map__,
                        type=field.type,
                    )
                    # fill the role manager
                    roles.by_num[role_num] = role_
                    roles.by_name[role_.name] = role_
                    roles.qt_roles[role_.num] = role_.qt_name

            cls.Model = GenericModel.from_role_defined(cls, parent=getattr(cls, "Model", None))
            cls.Model.__roles__ = roles

        return cls


class _BaseType(metaclass=BaseTypeMeta):
    if TYPE_CHECKING:  # pragma: no cover
        __types_map__: dict[str, _BaseType]
        Model: ClassVar[type[GenericModel[Self]]]  # type: ignore

    def asdict(self) -> dict:
        return attrs.asdict(self)  # type: ignore


_TBaseType = TypeVar("_TBaseType", bound=_BaseType)


def get_base_type(name: str = "Default") -> Type[_BaseType]:
    """:return: BaseType to use for generating `GenericModel`'s"""
    cls: type[_BaseType] = types.new_class(name=f"{name}BaseType", kwds={"metaclass": BaseTypeMeta})
    # we need to inject this here in order for `get_type_hints()` to work.
    cls.__types_map__ = {GenericModel.__name__: GenericModel}  # type: ignore
    return cls


class Role:
    """Provides metadata about a field in a define_roles decorated class."""

    __slots__ = ("num", "name", "qt_name", "_type", "str_type", "type_map")

    def __init__(self, num: int, name: str, type: Union[TypeHinter, str], type_map: dict):
        self.num = num
        self.name = name
        self.qt_name = QByteArray(name.encode("utf-8"))
        self.type_map = type_map
        if isinstance(type, str):
            self.str_type = type
        else:
            self._type = TypeHinter.from_annotations(type)

    @property
    def type(self) -> TypeHinter:
        try:
            return self._type
        except AttributeError:
            self._type = TypeHinter.from_string(self.str_type, ns=self.type_map)
            return self._type


class RoleMapper:
    """A container that maps the roles of a defined class each map has a
    certain usage in the future."""

    def __init__(self) -> None:
        # this is the real name of the field
        # how the class would be created
        self.by_name: dict[str, Role] = {}
        # number is  a qt role 256+
        # this is how qml will call the data() method
        self.by_num: dict[int, Role] = {}
        # just to return to qt method roleNames()
        self.qt_roles: dict[int, QByteArray] = {}

    @cached_property
    def children(self) -> dict[str, type[_BaseType]]:
        # mapping all the roledefined to know for what
        # to initialize a genericModel
        ret = {}
        for role_ in self.by_name.values():
            # lists must be child models
            if role_.type.type is GenericModel:
                child_type = role_.type.of_type[0].type
                ret[role_.name] = child_type
        return ret


class RoleDoesNotExist(QtGqlException):
    ...


class GenericModel(Generic[_TBaseType], QAbstractListModel):
    """Contains logic autogenerated models that extends `_BaseType` You don't
    need to use this class directly.

    Although you would need it to annotate "child" roles. i.e:

    ```python
    from qtgql.tools.itemsystem import get_base_type, GenericModel

    BaseType = get_base_type()
    class SomeModel(BaseType):
         other_model: GenericModel['OtherModel']
    ```
    """

    __roledefined_type__: type[_TBaseType]
    __roles__: RoleMapper
    layoutAboutToBeChanged: Signal  # type: ignore[misc]
    layoutChanged: Signal  # type: ignore[misc]
    dataChanged: Callable[[QModelIndex, QModelIndex, Optional[list[int]]], None]  # type: ignore[assignment, misc]

    def __init__(self, *, data: Optional[Union[list[dict], dict]] = None, parent=None):
        super().__init__(parent)
        self.rowsAb = None

        self.parent_model: Optional[GenericModel] = parent
        self.type_ = self.__roledefined_type__
        self._data: list[_TBaseType] = []
        self._child_models: list[GenericModel[Any]] = []
        if data:
            self.initialize_data(data)

    def initialize_data(self, data: Union[list[dict], dict]) -> None:
        """:param data: data to generate the model"""
        self.layoutAboutToBeChanged.emit()
        QEventLoop().processEvents()
        # search for children and initialize them as models
        for node in data:
            if children := self.__roles__.children:
                for name, child in children.items():
                    child_data = node[name]
                    child_model = child.Model(data=child_data, parent=self)
                    node[name] = child_model
                    #  save child models for further use.
                    self._child_models.append(child_model)

        # initialize list of attrs classes
        for node in data:
            inst = self.type_(**node)  # type: ignore
            self._data.append(inst)

        self.layoutChanged.emit()

    def data(self, index, role=...):
        if index.row() < len(self._data) and index.isValid():
            try:
                return getattr(self._data[index.row()], self.__roles__.by_num[role].name, None)
            except KeyError as exc:
                if role in (
                    -1,
                    Qt.ItemDataRole.DisplayRole,
                    Qt.ItemDataRole.ToolTipRole,
                    Qt.ItemDataRole.StatusTipRole,
                    Qt.ItemDataRole.WhatsThisRole,
                    Qt.ItemDataRole.SizeHintRole,
                    Qt.ItemDataRole.FontRole,
                    Qt.ItemDataRole.BackgroundRole,
                    Qt.ItemDataRole.ForegroundRole,
                    Qt.ItemDataRole.DecorationRole,
                    Qt.ItemDataRole.TextAlignmentRole,
                    Qt.ItemDataRole.CheckStateRole,
                ):
                    return None
                # resolvers should be pre-evaluated when the model updated
                raise RoleDoesNotExist(
                    f"role {role} of type {self.type_} at index: [{index}] "
                    f"is not a valid role!\n"
                    f"options are: {self.__roles__.qt_roles}",
                ) from exc

    def setData(self, index, value, role=...):
        if index.row() < len(self._data) and index.isValid():
            setattr(self._data[index.row()], self.__roles__.by_num[role].name, value)
            return True
        return False

    def flags(self, index):
        if index.row() < len(self._data) and index.isValid():
            return Qt.ItemFlag.ItemIsEditable
        return Qt.ItemFlag.NoItemFlags

    def roleNames(self) -> dict:
        return self.__roles__.qt_roles

    def rowCount(self, parent=...) -> int:
        return len(self._data)

    def append(self, node: _TBaseType) -> None:
        count = self.rowCount()
        self.beginInsertRows(self.index(count), count, count)
        self._data.append(node)
        self.endInsertRows()

    @slot
    def pop(self, index: Optional[int] = None) -> None:
        index = -1 if index is None else index
        real_index = index if index > -1 else self.rowCount()
        self.beginRemoveRows(self.index(index - 1).parent(), real_index, real_index)
        self._data.pop(index)
        self.endRemoveRows()

    @slot
    def clear(self) -> None:
        if self._data:
            self.beginRemoveRows(self.index(0).parent(), 0, self.rowCount())
            for child in self._child_models:
                child.clear()
            self._data = []
            self.endRemoveRows()

    @slot
    def get_by_index(self, row: int, key: str) -> Any:
        return self.data(self.index(row), self.__roles__.by_name[key].num)

    # CLASS METHODS
    @classmethod
    def from_role_defined(
        cls,
        type_: type[_BaseType],
        parent: Optional[Type[GenericModel]] = None,
    ) -> type[GenericModel]:
        bases = (cls,)
        if parent:
            bases = (parent,)
        ret = type(cls.name_by_type(type_), bases, {"__roledefined_type__": type_})
        assert issubclass(ret, GenericModel)
        return ret

    @classmethod
    def name_by_type(cls, type_: type) -> str:
        return type_.__name__ + "Model"
